# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

import xml.etree.ElementTree as ET

from gentle.metadata import Person, RemoteID, MetadataXML

def test_person_to_xml():
    person = Person(name="Larry the Cow", email="larry@gentoo.org")
    assert ET.tostring(person.to_xml(), encoding="unicode") == '<maintainer><name>Larry the Cow</name><email>larry@gentoo.org</email></maintainer>'

    person = Person(name="Larry the Cow")
    assert ET.tostring(person.to_xml(), encoding="unicode") == '<maintainer><name>Larry the Cow</name></maintainer>'

    person = Person(email="larry@gentoo.org")
    assert ET.tostring(person.to_xml(), encoding="unicode") == '<maintainer><email>larry@gentoo.org</email></maintainer>'

def test_remote_id_to_xml():
    remote_id = RemoteID(attr="github", value="gentoo/gentoo")
    assert ET.tostring(remote_id.to_xml(), encoding="unicode") == '<remote-id type="github">gentoo/gentoo</remote-id>'



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2023 Anna <cyber@sysrq.in>
# No warranty

""" Generic metadata parsing structures and routines """

import logging
import textwrap
import xml.etree.ElementTree as ET
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import Optional

logger = logging.getLogger("metadata")


@dataclass
class Person:
    name: str = field(default="", compare=False)
    email: str = ""

    def to_xml(self, attrib: Optional[dict] = None) -> ET.Element:
        """
        :param attrib: attributes for the ``<maintainer>`` tag
        :return: :file:`metadata.xml` respresentation of a person
        """
        result = ET.Element("maintainer", attrib=attrib or {})
        if self.name:
            name_elem = ET.SubElement(result, "name")
            name_elem.text = self.name
        if self.email:
            email_elem = ET.SubElement(result, "email")
            email_elem.text = self.email

        return result


@dataclass
class RemoteID:
    attr: str
    value: str

    def to_xml(self) -> ET.Element:
        """
        :return: :file:`metadata.xml` respresentation of a remote id
        """
        remote_elem = ET.Element("remote-id", type=self.attr)
        remote_elem.text = self.value
        return remote_elem


@dataclass
class Upstream:
    maintainers: list[Person] = field(default_factory=list)
    changelog: Optional[str] = None
    doc: Optional[str] = None
    bugs_to: Optional[str] = None
    remote_ids: list[RemoteID] = field(default_factory=list)


class MetadataXML:
    """ Parse and write :file:`metadata.xml` files. """

    def __init__(self, xmlfile: Path):
        """
        :param xmlfile: path to the :file:`metadata.xml` file
        """
        self.xmlfile: Path = xmlfile
        self.xml: ET = ET.parse(self.xmlfile)

        self._maintainers: list[Person] = []
        self._upstream = Upstream()
        self._parse_metadata_xml()

    @property
    def maintainers(self) -> list[Person]:
        """ List of package maintainers """
        return self._maintainers.copy()

    @property
    def upstream(self) -> Upstream:
        """ Upstream information """
        return replace(self._upstream)

    def dump(self) -> None:
        """ Write :file:`metadata.xml` file """
        logger.info("Writing metadata.xml")
        ET.indent(self.xml, space="\t", level=0)
        with open(self.xmlfile, "w") as file:
            file.write(textwrap.dedent("""\
                <?xml version="1.0" encoding="UTF-8"?>
                <!DOCTYPE pkgmetadata SYSTEM "https://www.gentoo.org/dtd/metadata.dtd">
            """))
            self.xml.write(file, encoding="unicode")
            file.write("\n")

    def dumps(self) -> None:
        """ Convert the object to text """
        ET.indent(self.xml, space="\t", level=0)
        return ET.tostring(self.xml.getroot(), encoding="unicode")

    def add_upstream_maintainer(self, person: Person) -> None:
        """ Add a person to the list of upstream maintainers """
        if person in self._upstream.maintainers:
            return

        logger.info("Adding upstream maintainer: %s", person)
        self._upstream.maintainers.append(person)
        upstream = self._make_upstream_element()
        upstream.append(person.to_xml())

    def add_upstream_remote_id(self, remote_id: RemoteID) -> None:
        """ Add an item to the list of remote ids """
        if remote_id in self._upstream.remote_ids:
            return

        logger.info("Adding remote id: %s", remote_id)
        self._upstream.remote_ids.append(remote_id)
        upstream = self._make_upstream_element()
        upstream.append(remote_id.to_xml())

    def set_upstream_bugs_to(self, url: str) -> None:
        """ Set upstream bugs-to URL """
        if self._upstream.bugs_to:
            return

        logger.info("Setting upstream bug tracker to %s", url)
        self._upstream.bugs_to = url

        upstream = self._make_upstream_element()
        bugs_to = ET.SubElement(upstream, "bugs-to")
        bugs_to.text = url

    def set_upstream_changelog(self, url: str) -> None:
        """ Set upstream changelog URL """
        if self._upstream.changelog:
            return

        logger.info("Setting upstream changelog to %s", url)
        self._upstream.changelog = url

        upstream = self._make_upstream_element()
        changelog = ET.SubElement(upstream, "changelog")
        changelog.text = url

    def set_upstream_doc(self, url: str) -> None:
        """ Set upstream documentation URL """
        if self._upstream.doc:
            return

        logger.info("Setting upstream documentation to %s", url)
        self._upstream.doc = url

        upstream = self._make_upstream_element()
        doc = ET.SubElement(upstream, "doc")
        doc.text = url

    def _make_upstream_element(self) -> ET.Element:
        if (upstream := self.xml.find("upstream")) is None:
            pkgmetadata = self.xml.getroot()
            upstream = ET.SubElement(pkgmetadata, "upstream")
        return upstream

    def _parse_metadata_xml(self) -> None:
        """ Parse the :file:`metadata.xml` file """
        assert self.xml.getroot().tag == "pkgmetadata"

        for maint in self.xml.findall("maintainer"):
            self._parse_package_maintainer(maint)

        if (upstream := self.xml.find("upstream")) is not None:
            self._parse_upstream(upstream)

    def _parse_person(self, xml: ET.Element, *, name_optional: bool = False,
                      email_optional: bool = False) -> Optional[Person]:
        """
        :param xml: ``<maintainer>`` XML tag
        """
        assert xml.tag == "maintainer"

        if (email_tag := xml.find("email")) is not None:
            email = "".join(email_tag.itertext())
        elif email_optional:
            email = None
        else:
            return None

        if (name_tag := xml.find("name")) is not None:
            name = "".join(name_tag.itertext())
        elif name_optional:
            name = None
        else:
            return None

        return Person(name, email)

    def _parse_package_maintainer(self, xml: ET.Element) -> None:
        """
        :param xml: top-level ``<maintainer>`` XML tag
        """
        if (person := self._parse_person(xml, name_optional=True)) is not None:
            self._maintainers.append(person)

    def _parse_upstream(self, xml: ET.Element) -> None:
        """
        :param xml: ``<upstream>`` XML tag
        """
        assert xml.tag == "upstream"

        for maint in xml.findall("maintainer"):
            if (person := self._parse_person(maint, email_optional=True)) is not None:
                self._upstream.maintainers.append(person)

        for remote in xml.findall("remote-id"):
            attr = remote.get("type")
            value = "".join(remote.itertext())
            self._upstream.remote_ids.append(RemoteID(attr, value))

        if (changelog := xml.find("changelog")) is not None:
            self._upstream.changelog = "".join(changelog.itertext())

        if (doc := xml.find("doc")) is not None:
            self._upstream.doc = "".join(doc.itertext())

        if (bugs_to := xml.find("bugs-to")) is not None:
            self._upstream.bugs_to = "".join(bugs_to.itertext())



# SPDX-License-Identifier: WTFPL
# SPDX-FileCopyrightText: 2022-2023 Anna <cyber@sysrq.in>
# No warranty

import argparse
import logging
import os
import re
from pathlib import Path
from tempfile import TemporaryDirectory

import gentle
from gentle.generators import AbstractGenerator
from gentle.metadata import MetadataXML

import gentle.generators.bower
import gentle.generators.cargo
import gentle.generators.composer
import gentle.generators.npm
import gentle.generators.pyproject
import gentle.generators.shards

try:
    import portage
    from portage.package.ebuild.doebuild import doebuild
except ModuleNotFoundError:
    _HAS_PORTAGE = False
else:
    _HAS_PORTAGE = True

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("main")


def portage_src_unpack(ebuild: Path, tmpdir: str) -> Path:
    """
    Unpack the sources using Portage.

    :param ebuild: path to the ebuild file
    :param tmpdir: temporary directory
    :return: the value of ``${S}``
    """
    ebuild = ebuild.resolve()
    portdir = str(ebuild.parents[2])

    # pylint: disable=protected-access
    if portdir not in portage.portdb.porttrees:
        portdir_overlay = portage.settings.get("PORTDIR_OVERLAY", "")
        os.environ["PORTDIR_OVERLAY"] = portdir_overlay + " " + portage._shell_quote(portdir)

        print(f"Appending {portdir} to PORTDIR_OVERLAY...")
        portage._reset_legacy_globals()

    tmpsettings: portage.config = portage.portdb.doebuild_settings
    tmpsettings["PORTAGE_USERNAME"] = os.getlogin()
    tmpsettings["PORTAGE_TMPDIR"] = tmpdir
    tmpsettings["DISTDIR"] = tmpdir
    tmpsettings.features._features.clear()  # pylint: disable=protected-access
    tmpsettings.features.add("unprivileged")
    settings = portage.config(clone=tmpsettings)

    status = doebuild(str(ebuild), "unpack",
                      tree="porttree",
                      settings=settings,
                      vartree=portage.db[portage.root]["vartree"])
    if status != 0:
        raise RuntimeError("Unpack failed")

    env = Path(settings.get("T")) / "environment"
    srcdir_re = re.compile(r'^declare -x S="(?P<val>.+)"$')
    with open(env) as file:
        for line in file:
            if (match := srcdir_re.match(line)) is not None:
                return Path(match.group("val"))
    raise RuntimeError("No ${S} value found")


def main() -> None:
    """
    Parse command-line arguments and run the program.
    """

    pm = []
    if _HAS_PORTAGE:
        pm.append("portage")

    if len(pm) == 0:
        raise RuntimeError("No package manager installed. Aborting")

    parser = argparse.ArgumentParser("gentle", description=gentle.__doc__)
    parser.add_argument("ebuild", type=Path, help="path to the ebuild file")
    parser.add_argument("--api", "-a", choices=pm, default=pm[0],
                        help="package manager API to use")
    parser.add_argument("-v", action="version", version=gentle.__version__)
    args = parser.parse_args()

    mxml_file = args.ebuild.parent / "metadata.xml"
    mxml = MetadataXML(mxml_file)

    with TemporaryDirectory(prefix="gentle-") as tmpdir:
        srcdir = portage_src_unpack(args.ebuild, tmpdir)
        for cls in AbstractGenerator.get_generator_subclasses():
            generator = cls(srcdir)
            if generator.active:
                logger.info("Starting %s", cls.__name__)
                generator.update_metadata_xml(mxml)
    mxml.dump()


if __name__ == "__main__":
    main()
